import pickle
import queue
import random
import threading

from lol_scraper.concurrency import worker
from lol_scraper.concurrency.common import close_queue
from lol_scraper.data_types import ConcurrentSet, QueueMultiplexer, DistributedLogger, FallbackQueue, Bunch
from lol_scraper.summoners_api import get_patch_changed, LATEST, consume_path_changed

MAX_LOCAL_QUEUE_SIZE = 50
EVICTION_SIZE = 50000
EVICTION_RATE = 0.2

# Token fields
TYPE = 0
PAYLOAD = 1

# Token recipients
MAIN_ID = -1
ANY = -2

# Token types
EXIT_REQUEST = 1
PAST_PLAYERS_AND_MATCHES = 2
INITIATING_SHUTDOWN = 3
SHUTDOWN_COMPLETED = 4
MONITORING_REQUEST = 5
MONITORING_RESPONSE = 6


def msg_ignore(*types):
    return lambda message: get_type(message) in types


def make_message(type, recipient, payload=None):
    return (type, payload)


def get_payload(message):
    return message[PAYLOAD]


def get_type(message):
    return message[TYPE]


# TODO change the queue.put() to use the timeout argument so that the process never blocks for too much


def _initialize_past_sets(bucket_id, main_connection, past_players, past_matches, logger):
    retries = 3
    while retries:
        try:
            _message = main_connection.get(timeout=1)
        except queue.Empty:
            retries -= 1
        else:
            if get_type(_message) == PAST_PLAYERS_AND_MATCHES:
                initial_past_players, initial_past_matches = get_payload(_message)
                past_players.update(initial_past_players)
                past_matches.update(initial_past_matches)
                return
    # retries got to 0. Notify that we didn't get the messages with the initial players
    logger.info("No player initialization messages where received in 3 tries. Continuing with empty set")


def _evict_players(past_players, logger):
    # past_players grows indefinitely. This doesn't make sense, as after a while a player have new
    # matches. When the list grows too big we remove part of the players, so that they can be analyzed again
    if len(past_players) > EVICTION_SIZE:
        total_players_pre_eviction = len(past_players)
        # Store the players to remove. We need to do this to be able to operate on the past_player object
        # directly, without changing it's reference, as the threads have an handle to the reference
        players_to_evict = {p_id for p_id in past_players if random.random() > EVICTION_RATE}
        past_players.difference_update(players_to_evict)
        msg = "Evicting analyzed players. Previously: {}. Now: {}".format(total_players_pre_eviction,
                                                                          len(past_players))
        logger.info(msg)


def _clear_past_matches(conf, past_matches, logger):
    # When a new patch is released, we can clear all the downloaded_matches if minimum_patch == 'latest'
    # Most of the time it will be False: do not acquire the lock in that case
    if get_patch_changed() and conf['minimum_patch'].lower() == LATEST:
        past_matches.clear()
        consume_path_changed()
        logger.info("New patch detected. Cleaned the downloaded matches set")


def download_and_execute(conf, bucket_id, mp_player_queues, mp_match_queues, mp_downloaded_matches, main_connection,
                         user_function, thread_init_function, logging_level, logging_queue,
                         initialization_function, number_of_threads):
    logger = DistributedLogger(logging_level, logging_queue, prologue="[Bucket {}] ".format(bucket_id))
    started_threads = []
    local_exit_request = Bunch(value=False)
    past_matches = None
    past_players = None
    next_player = None
    next_match = None
    downloaded_matches = None
    try:
        # The settings aren't transferred from one process to the other. Call an initialization function which should set
        # the state to what is needed for the execution
        init_func = pickle.loads(initialization_function)
        user_data = init_func(bucket_id, conf)

        # Initialize the queues to pass to the workers
        next_player = QueueMultiplexer(bucket_id, queue.Queue(maxsize=MAX_LOCAL_QUEUE_SIZE), mp_player_queues)
        next_match = QueueMultiplexer(bucket_id, queue.Queue(maxsize=MAX_LOCAL_QUEUE_SIZE), mp_match_queues)
        downloaded_matches = FallbackQueue(queue.Queue(maxsize=MAX_LOCAL_QUEUE_SIZE), mp_downloaded_matches)

        past_players = ConcurrentSet()
        past_matches = ConcurrentSet()
        # Initialize the past sets with the previous elements. Retry 5 times, after that continue, but notify the main
        _initialize_past_sets(bucket_id, main_connection, past_players, past_matches, logger)

        assert len(number_of_threads) == 3
        assert all(x >= 0 for x in number_of_threads)
        # TODO a smarter, automatic, dynamic evaluation of the number of threads to use is possible. Do it eventually
        num_player_threads, num_match_threads, num_user_threads = number_of_threads
        # Create and start the workers
        for i in range(num_player_threads):
            player_downloader = threading.Thread(name="PlayerDownloader%s" % i, target=worker.download_players, args=(
                conf, next_player, next_match, past_players, logger, local_exit_request))
            player_downloader.start()
            started_threads.append(player_downloader)
        logger.debug("Started Player Downloaders")

        for i in range(num_match_threads):
            match_downloader = threading.Thread(name="MatchDownloader%s" % i, target=worker.download_matches, args=(
                conf, next_match, next_player, downloaded_matches, past_matches, logger, local_exit_request))
            match_downloader.start()
            started_threads.append(match_downloader)
        logger.debug("Started Match Downloaders")

        user_func = pickle.loads(user_function)
        thread_init = pickle.loads(thread_init_function)
        for i in range(num_user_threads):
            user_func_executor = threading.Thread(name="UserFunctionExecutor%s" % i, target=worker.function_executor,
                                                  args=(user_func, downloaded_matches, thread_init, i, user_data,
                                                        logger, local_exit_request))
            user_func_executor.start()
            started_threads.append(user_func_executor)
        logger.debug("Started User Function Executor")

        # Monitor the execution
        while True:
            try:
                # Format: (type, payload)
                message = main_connection.get(timeout=10)
            except queue.Empty:
                pass
            else:
                if get_type(message) == EXIT_REQUEST:
                    logger.debug("Exit requested")
                    # Run the finally code and exit
                    return
                elif get_type(message) == MONITORING_REQUEST:
                    request_id = get_payload(message)
                    logger.debug("Received monitoring request %i", request_id)
                    payload = (
                        next_player.qsize(),
                        len(past_players),
                        next_match.qsize(),
                        len(past_matches),
                        downloaded_matches.qsize(local_only=True)
                    )
                    main_connection.put(make_message(MONITORING_RESPONSE, MAIN_ID, (request_id, payload)))
                    logger.debug("Sent monitoring response")
                else:
                    raise ValueError("Unexpected message: {}".format(message))

            _evict_players(past_players, logger)
            _clear_past_matches(conf, past_matches, logger)

    except:
        logger.exception("Exception in the controller of the process")
        raise
    finally:
        # We want to log the exceptions that might happen here
        try:
            logger.debug("Initiating shutdown")
            main_connection.put(make_message(INITIATING_SHUTDOWN, MAIN_ID))
            # TODO we might receive a EXIT_REQUEST if it was us who initiated the termination. Should we consume it?
            # Notify the threads to exit
            local_exit_request.value = True

            # Wait fot the threads to terminate
            for thread in started_threads:
                thread.join()
            logger.debug("Joined the threads")

            # Put in the multiprocess queues the data which was stored locally, so the main can inspect them
            main_connection.put(make_message(PAST_PLAYERS_AND_MATCHES, MAIN_ID,
                                             (past_players or set(), past_matches or set())))
            logger.debug("Sent the past player and matches")

            # Put the items in the local queues in the multiprocess queues. We need to be sure not to access the
            # multiplexer if it was not assigned before an exception brought us here
            # Bunch.get() will always raise a queue.Empty exception
            empty_queue = Bunch(get=lambda *args, **kwargs: (_ for _ in ()).throw(queue.Empty))
            for multiplexer, remote_queues in ((next_player or empty_queue, mp_player_queues),
                                               (next_match or empty_queue, mp_match_queues)):
                while True:
                    try:
                        player = multiplexer.get(block=False, weight=1)
                        remote_queues[bucket_id].put(player)
                    except queue.Empty:
                        break
            if downloaded_matches:
                downloaded_matches.flush_to_fallback()
            logger.debug("Flushed local queues to the multiprocess ones")

            # bucket 0 must not close the queues, as the main is still using them to join the other processes
            if bucket_id != 0:
                queues_to_close = mp_player_queues + mp_match_queues + [mp_downloaded_matches]
                close_queue(*queues_to_close)
                logger.debug("Closed the queues")
            main_connection.put(make_message(SHUTDOWN_COMPLETED, MAIN_ID, None))
        except:
            logger.exception("Exception while terminating the controller")
            raise
        else:
            logger.debug("Terminated successfully")
            logger.debug("Terminated successfully")


import threading
import datetime
import queue
import random
import logging
import sys
import time

import time as _time
from collections import deque
from enum import Enum, unique

from lol_scraper.concurrency.common import exc2pickle, task_done


@unique
class Queue(Enum):
    RANKED_SOLO_5x5 = 0
    RANKED_TEAM_3x3 = 1
    RANKED_TEAM_5x5 = 2


@unique
class Tier(Enum):
    challenger = 0
    master = 1
    diamond = 2
    platinum = 3
    gold = 4
    silver = 5
    bronze = 6

    @classmethod
    def equals_and_above(cls, tier):
        for t in cls:
            if t.is_better_or_equal(tier):
                yield t

    @classmethod
    def all_tiers_below(cls, tier):
        for t in cls:
            if not t.is_better_or_equal(tier):
                yield t

    def __hash__(self):
        return self.value

    def __eq__(self, other):
        return hasattr(other, "value") and self.value == other.value

    def best(self, other):
        if self.value <= other.value:
            return self
        return other

    def worst(self, other):
        if self.value >= other.value:
            return self
        return other

    def is_better_or_equal(self, other):
        return self.value <= other.value

    @classmethod
    def parse(cls, tier):
        initial = tier[0].lower()
        if initial == 'b':
            return cls.bronze
        elif initial == 's':
            return cls.silver
        elif initial == 'g':
            return cls.gold
        elif initial == 'p':
            return cls.platinum
        elif initial == 'd':
            return cls.diamond
        elif initial == 'm':
            return cls.master
        elif initial == 'c':
            return cls.challenger
        else:
            raise ValueError("No Tier with name {}".format(tier))


@unique
class Maps(Enum):
    SUMMONERS_RIFT = 11


class SimpleCache():
    def __init__(self):
        self.store = {}

    def set(self, key, value, time=0):
        self.store[key] = (value, time, _time.time())

    def get(self, key, default=None):
        item = self.store.get(key, None)
        if item is None:
            return default
        else:
            value, time, put_time = item
            current_time = _time.time()
            # time = 0 means it never expires
            if put_time + time > current_time or time == 0:
                # it is still valid
                return value
            else:
                # expired value
                del self.store[key]
                return default


def cache_autostore(key, duration, cache, args_to_str=None, on_change=None):
    def make_key(*args, **kwargs):
        if args_to_str:
            return str(key) + args_to_str(*args, **kwargs)
        else:
            return key

    def function_decorator(wrapped):
        def wrapper(*args, **kwargs):
            sentinel = object()
            composed_key = make_key(*args, **kwargs)
            # get the value
            cached_value = cache.get(composed_key, sentinel)

            # if the value is saved or expired
            if cached_value is sentinel:
                # call the function which gives the real value
                new_value = wrapped(*args, **kwargs)
                # store it
                cache.set(composed_key, new_value, duration)
                # if we want to be notified of the change
                if on_change:
                    # get the old value
                    old = cache.get(composed_key + "_old")
                    # set the new value
                    cache.set(composed_key + "_old", new_value, 0)
                    if old != new_value:
                        on_change(old, new_value)
                # Then return the new value
                return new_value
            else:
                # The value stored was still fresh, return it
                return cached_value

        return wrapper

    return function_decorator


epoch = datetime.datetime.utcfromtimestamp(0)


def milliseconds_unix_time(dt=None):
    """
    Compute the milliseconds elapsed from the Unix Epoch (1/01/1970) to a specific date
    :param dt: A date. If it's equal to None, datetime.now() is used
    :return: The milliseconds since the Unix Epoch
    """
    if dt is None:
        dt = datetime.datetime.now()
    return int(unix_time(dt) * 1000)


def unix_time(dt):
    """
    Compute the seconds elapsed from the Unix Epoch (1/01/1970) to a specific date
    :param dt: A date
    :return: The seconds since the Unix Epoch
    """
    delta = dt - epoch
    return delta.total_seconds()


def do_every(seconds, func=None, *args, **kwargs):
    def g_tick():
        t = time.time()
        count = 0
        while True:
            count += 1
            yield max(t + count * seconds - time.time(), 0)

    g = g_tick()
    while True:
        time.sleep(next(g))
        # This allows to use it in a for loop with a loop every 'seconds' seconds
        if func is None:
            yield
        else:
            func(*args, **kwargs)


class ConcurrentSet(set):
    """
    A set wrapped in a lock. THIS SET IS NOT SYNCHRONIZED AUTOMATICALLY, but makes it easier to use it as a synchronized
    set.
    To access it in a synchronized way use the context manager syntax:

    with this_set:
        this_set.add(item)
        this_set.pop()
        this_set.clear()

    """

    def __init__(self, seq=(), lock=None):
        super(ConcurrentSet, self).__init__(seq)
        if lock:
            if not hasattr(lock, '__enter__') and hasattr(lock, '__exit__'):
                raise ValueError("The lock must implement the context manager interface")
            self.lock = lock
        else:
            self.lock = threading.RLock()

    def __enter__(self):
        return self.lock.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.lock.__exit__(exc_type, exc_val, exc_tb)


class QueueMultiplexer:
    def __init__(self, bucket_id, local_queue, remote_queues, alpha=0.1):
        self.bucket_id = bucket_id
        self.local_q = local_queue
        self.remote_q = remote_queues
        self.num_partitions = len(remote_queues)
        self.weight = (1 + alpha )/ (self.num_partitions + alpha)

        assert bucket_id < self.num_partitions

        # When num_partition weight == 1. This prevents to get elements from the remote queue.
        # The behaviour is correct: since there is only one bucket the data will always be put in the local queue,
        # but the remote queue might contain some initial data when the multiplexer is created. We move them to the
        # local queue
        if self.num_partitions == 1:
            self._move_to_local_queue()

    def _move_to_local_queue(self):
        while True:
            try:
                self.local_q.put(self.remote_q[self.bucket_id].get(block=False))
                task_done(self.remote_q[self.bucket_id])
            except queue.Empty:
                break
            except queue.Full:
                break

    def get(self, block=True, timeout=None, weight=None):
        """
        Try getting an item from a queue. The queue is either the local queue or the remote queue with the id equal
        to this multiplexer bucket_id. The queue is chosen randomly, with weight defining the probability of picking
        the local queue.
        If the selected queue is empty, try to get(block=False) from the other queue (to not sleep more than
        the specified timeout)
        :param block:       See queue.Queue.get documentation
        :param timeout:     See queue.Queue.get documentation
        :param weight:      The probability to pick the local queue to perform the get operation.
                            Defaults to (1 + alpha) / (total_buckets + alpha)
        :return:            See queue.Queue.get documentation
        """
        if weight is None:
            weight = self.weight
        if random.random() < weight:
            first = self.local_q
            second = self.remote_q[self.bucket_id]
        else:
            second = self.local_q
            first = self.remote_q[self.bucket_id]

        try:
            item = first.get(block=False)
            task_done(first)
            return item
        except queue.Empty:
            item = second.get(block, timeout)
            task_done(second)
            return item

    def put(self, item, block=True, timeout=None):
        destination_bucket = item % self.num_partitions
        if destination_bucket == self.bucket_id:
            destination_queue = self.local_q
        else:
            destination_queue = self.remote_q[destination_bucket]

        destination_queue.put(item, block=block, timeout=timeout)

    def qsize(self, bucket=None):
        """
        Returns the sum of the elements in the local queue and in the queue specified by bucket
        :param bucket: The remote queue size to add to the local queue. Defaults to this multiplexer bucket_id
        :return: The size of the local queue plus the size of the remote queue at id equal to bucket
        """
        if bucket is None:
            bucket = self.bucket_id
        return self.local_q.qsize() + self.remote_q[bucket].qsize()


class FallbackQueue:
    def __init__(self, main_queue, fallback_queue):
        self.main_queue = main_queue
        self.fallback_queue = fallback_queue

    def get(self, block=True, timeout=None, **kwargs):
        # kwargs is used in order to provide the same interface as QueueMultiplexer
        try:
            item = self.main_queue.get(block=False)
            task_done(self.main_queue)
            return item
        except queue.Empty:
            item = self.fallback_queue.get(block=block, timeout=timeout)
            task_done(self.fallback_queue)
            return item

    def put(self, item, block=True, timeout=None):
        try:
            self.main_queue.put(item, block=False)
        except queue.Full:
            self.fallback_queue.put(item, block=block, timeout=timeout)

    def qsize(self, local_only=False):
        if local_only:
            return self.main_queue.qsize()
        else:
            return self.main_queue.qsize() + self.fallback_queue.qsize()

    def flush_to_fallback(self, block=True, timeout=None):
        """
        Moves all the items from the main queue to the fallback queue
        """
        while True:
            try:
                item = self.main_queue.get(block=False)
                task_done(self.main_queue)
                self.fallback_queue.put(item, block, timeout)
            except queue.Empty:
                return


class DistributedLogger():
    def __init__(self, level, queue, prologue=""):
        self._level = level
        self._queue = queue
        self._closed = False
        self._prologue = prologue

    @property
    def level(self):
        return self._level

    def getEffectiveLevel(self):
        return self._level

    def isEnabledFor(self, level):
        return level >= self._level

    def _log(self, level, message, *args, exc_info=None, exception=None, **kwargs):
        if self._closed:
            raise IOError("The logger is closed. You can't use it anymore")
        msg = self._prologue + (message % args)
        if exc_info:
            if exception:
                exc_info = (type(exception), exception, exception.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()
            # else exc_info is a tuple and we use it as is
            self._queue.put((level, msg, exc2pickle(exc_info)))
        else:
            self._queue.put((level, msg))

    def log(self, level, msg, *args, **kwargs):
        if not isinstance(level, int):
            raise TypeError("level must be an integer")
        if self.isEnabledFor(level):
            self._log(level, msg, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        if self.isEnabledFor(logging.DEBUG):
            self._log(logging.DEBUG, message, *args, **kwargs)

    def info(self, message, *args, **kwargs):
        if self.isEnabledFor(logging.INFO):
            self._log(logging.INFO, message, *args, **kwargs)

    def warn(self, message, *args, **kwargs):
        if self.isEnabledFor(logging.WARN):
            self._log(logging.WARN, message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        if self.isEnabledFor(logging.ERROR):
            self._log(logging.ERROR, message, *args, **kwargs)

    def exception(self, message="", *args, exc_info=True, **kwargs):
        kwargs['exc_info'] = exc_info
        self.error(message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        if self.isEnabledFor(logging.CRITICAL):
            self._log(logging.CRITICAL, message, *args, **kwargs)


class MessagePipe:
    def __init__(self, pipe):
        self.pipe = pipe

    def get(self, block=True, timeout=None):
        data_available = self.pipe.poll(timeout) if block else self.pipe.poll()
        if data_available:
            try:
                return self.pipe.recv()
            except EOFError:
                raise queue.Empty
        else:
            raise queue.Empty

    def put(self, item, block=True, timeout=None):
        try:
            self.pipe.send(item)
        except BrokenPipeError:
            raise queue.Full

    def get_ignore(self, ignore, block=True, timeout=None):
        while True:
            item = self.get(block, timeout)
            if ignore(item):
                continue
            return item


class Bunch:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)



import unittest
import os
from lol_scraper.summoners_api import summoner_names_to_id, _slice, leagues_by_summoner_ids, Tier
from cassiopeia import baseriotapi



class TierTest(unittest.TestCase):

    api_key = None
    def setUp(self):
        if not self.api_key:
            with open(os.path.join(os.path.dirname(__file__), '../../api-key'),'rt') as f:
                self.api_key = f.read()[:36]
        baseriotapi.print_calls(False)
        baseriotapi.set_region('euw')
        baseriotapi.set_api_key(self.api_key)

    def list_to_tier_initializer(self, list):
        initializer = {}
        for tier in Tier:
            if list[tier.value]:
                initializer[tier] = set(list[tier.value])
        return initializer

    def test_slice_empty_list(self):
        for _ in _slice(0,0, 1):
            self.fail()

    def test_slice_middle_values(self):
        source = [x for x in range(50)]
        lst = [(x, x+1) for x in range(50)]
        for (b1,e1), (b2,e2) in zip(lst, _slice(0,50,1)):
            self.assertEqual(source[b1:e1], source[b2:e2])

    def test_slice_end_value(self):
        step = 3
        lst = [ x for x in range(20)]
        begin = [0, 3, 6, 9,12,15,18]
        end =   [3, 6, 9,12,15,18,20]
        generator = _slice(0,20,step)
        for (b1,e1), (b2,e2) in zip(zip(begin, end), generator):
            self.assertEqual(lst[b1:e1], lst[b2:e2])
        try:
            next(generator)
            self.fail()
        except StopIteration:
            self.assertTrue(True)

    def test_leagues_by_summoner(self):
        summoners = [44256841, 21653685, 22447540, 22281234, 29378330, 23836705, 23746128, 23742827, 35473944,
                     56067323, 29600081, 26191711]
        leagues = {
                    Tier.challenger : {44256841, 21653685, 22447540},
                    Tier.master : {22281234},
                    Tier.diamond : {29378330, 23746128, 23742827, 23836705, 23836705},
                    Tier.platinum : {29600081},
                    Tier.gold : {35473944},
                    Tier.silver :{56067323},
                    Tier.bronze : {26191711}
                }
        result = leagues_by_summoner_ids(summoners)
        for tier in Tier:
            self.assertEqual(leagues[tier], result.get(tier, set()), tier.name)

    def test_names_to_id(self):
        names = ['cwfreeze', 'makersf', 'zoffo', 'w4sh', "sirnukesalot", "exngodzukee",
                 "lamept", "hiddenlaw", "dijio", "zockchock", "sirkillerlord"]

        ids = {'cwfreeze':44256841, 'makersf':45248415, 'zoffo':46677170,
               'w4sh':23367874, 'exngodzukee': 22447540, 'sirnukesalot': 21653685, 'dijio': 23742827,
               'hiddenlaw': 23746128,'lamept': 23836705,'sirkillerlord': 26191711,'zockchock': 56067323}
        result = summoner_names_to_id(names)
        self.assertEqual(ids, result)

if __name__ == '__main__':
    unittest.main()



"""
# Transform this to an example on how to use the library

import os
import argparse
import logging
import pickle
from json import loads
from contextlib import closing, suppress

from lol_scraper.persist import TierStore
from lol_scraper.match_downloader import setup_riot_api, prepare_config, download_matches

current_state_extension = '.pickle'


def make_store_callback(store):
    def store_callback(match, tier):
        store.store(match.to_json(sort_keys=False,indent=None), tier)
    return store_callback


def download_from_config(conf, store_callback, checkpoint_callback):
    setup_riot_api(conf)
    runtime_config = prepare_config(conf)
    download_matches(store_callback, checkpoint_callback, runtime_config)


def time_slice_end_callback(config_file, players_to_analyze, analyzed_players, matches_to_download, downloaded_matches):
    with open(config_file + current_state_extension, mode='wb') as matches:
        pickle.dump((players_to_analyze, matches_to_download, downloaded_matches), matches)


def load_players_and_matches_ids_into(config_file, conf):
    with suppress(FileNotFoundError), open(config_file + current_state_extension, mode='rb') as matches:
        players_to_analyse, matches_to_download, downloaded_matches = pickle.load(matches)
        conf['seed_players_id'] = players_to_analyse
        conf['matches_to_download_by_tier'] = matches_to_download
        conf['downloaded_matches'] = downloaded_matches


def main(configuration_file, no_state=False):
    with open(configuration_file, 'rt') as config_file:
        json_conf = loads(config_file.read())

    load_players_and_matches_ids_into(configuration_file, json_conf)

    base_file_name = json_conf.get('base_file_name', '')
    matches_per_file = json_conf.get('matches_per_file', 0)
    destination_directory = json_conf['destination_directory']
    # Allow the directory to be relative to the config file.
    if destination_directory.startswith('__file__'):
        configuration_file_dir = os.path.dirname(os.path.realpath(configuration_file))
        destination_directory=destination_directory.replace('__file__', configuration_file_dir)

    checkpoint_callback = lambda *args, **kwargs: time_slice_end_callback(configuration_file, *args, **kwargs) if not no_state else None

    with closing(TierStore(destination_directory, matches_per_file, base_file_name)) as store:
        download_from_config(json_conf, make_store_callback(store), checkpoint_callback)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('configuration_file',help='The json file to hold the configuration of the download session '
                                                  'you want to start by running this script. Might be a file saved '
                                                  'from a previous session',action='store')
    parser.add_argument('--no-state', action='store_true', help='Do not store in a .state file the current state of '
                                                           'execution, so that if the process is stopped it can be '
                                                           'resumed from the last state saved',
                        default=False)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s, %(levelname)s, %(name)s, %(message)s',
                        datefmt="%m-%d %H:%M:%S",
                        level=logging.INFO)

    main(args.configuration_file, args.no_state)
"""



import queue
import logging

from urllib.error import URLError
from lol_scraper.cassiopeia_proxy import APIError

from lol_scraper.concurrency.common import close_queue, pickle2exc

def _handle_exception(e, logger):
    if isinstance(e, APIError):
        if 400 <= e.error_code < 500:
            # Might be a connection problem
            logger.warning("Encountered error {}".format(e))
        elif 500 <= e.error_code < 600:
            # Server problem.
            logger.warning("Encountered error {}".format(e))
        else:
            logger.error("Encountered error {}".format(e))
    elif isinstance(e, URLError):
        logger.error("Encountered error {}. You are having connection issues".format(e))
    else:
        exc_info = (type(e), e, e.__traceback__)
        logger.error("Encountered unexpected exception {}".format(e), exc_info=exc_info)


def log_messages(logger, message_queue, exit_request):
    """
    :param logging.Logger           logger:
    :param queue.Queue              message_queue:
    :param multiprocessing.Value    exit_request:
    :return:
    """
    task_done = getattr(message_queue, 'task_done', lambda: None)
    while True:
        try:
            try:
                message = message_queue.get(timeout=0.5)
            except queue.Empty:
                # Only check for exit if there are no messages
                if exit_request.value:
                    break
                else:
                    continue
            else:
                task_done()

            level = message[0]
            msg = message[1]
            if len(message) == 2:
                logger.log(level, msg)
            elif len(message) == 3:
                exception = pickle2exc(message[2])
                _handle_exception(exception, logger)
        except Exception as e:
            _handle_exception(e, logger)
    close_queue(message_queue)


import datetime
import logging
import itertools
import urllib.error

from lol_scraper.cassiopeia_proxy import APIError, get_challenger, get_master, get_match_list

from lol_scraper.data_types import Tier, Queue, Maps, milliseconds_unix_time
from lol_scraper.summoners_api import summoner_names_to_id

delta_30_days = datetime.timedelta(days=30)

def prepare_config(config):
    runtime_config = {
        'cassiopeia': config['cassiopeia'],
        'logging_level': logging._nameToLevel[config.get('logging_level', 'NOTSET')],
        # Parse the time boundaries
        'end': None if 'end_time' not in config else datetime.datetime(**config['end_time']),
        'start': None if 'start_time' not in config else datetime.datetime(**config['start_time']),

        'minimum_patch': config.get('minimum_patch', ""),
        'queue': config.get('queue', Queue.RANKED_SOLO_5x5.name),
        'map_type': config.get('map', Maps.SUMMONERS_RIFT.name),
        'minimum_tier': config.get('minimum_tier', Tier.bronze.name).lower(),
        'include_timeline': config.get('include_timeline', True),

        'downloaded_matches': config.get('downloaded_matches', ()),
        'matches_to_download': config.get('matches_to_download', set()),
        'seed_players_id': config.get('seed_players_id', set())
    }

    if runtime_config['start'] is None:
        runtime_config['start'] = (runtime_config['end'] if runtime_config[
            'end'] else datetime.datetime.now()) - delta_30_days

    logger = logging.getLogger(__name__)
    if not runtime_config['seed_players_id']:
        # No players nor matches where provided, so we use the challenger and master players matches
        while True:
            try:
                config_seed_players = config.get('seed_players', None)
                if config_seed_players is None:
                    logger.info("No seed players detected. Downloading challenger and master players IDs.")
                    # Let's use challenger and master tier players as seed
                    runtime_config['seed_players_id'].update(int(league_entry_dto.playerOrTeamId) for league_entry_dto in
                                                        itertools.chain(get_challenger(runtime_config['queue']).entries,
                                                                        get_master(runtime_config['queue']).entries))
                else:
                    logger.info("Detected seed players names. Fetching the players IDs")
                    # We have a list of seed players. Let's use it
                    runtime_config['seed_players_id'].update(summoner_names_to_id(config_seed_players).values())

                break
            except APIError:
                # sometimes the network might have problems during the start. We don't want to crash just
                # because of that. Keep trying!
                logger.exception("APIError while initializing the script. Retrying")

    # We have the players ID, download the matches
    logger.info("Fetching up to 50 matches from the {} provided players".format(len(runtime_config['seed_players_id'])))
    while runtime_config['seed_players_id'] and len(runtime_config['matches_to_download']) < 50:
        player_id = runtime_config['seed_players_id'].pop()
        try:
            match_list = get_match_list(player_id, begin_time=milliseconds_unix_time(runtime_config['start']),
                            end_time=milliseconds_unix_time(runtime_config['end']),
                            ranked_queues=runtime_config['queue'])
            runtime_config['matches_to_download'].update(match.matchId for match in match_list.matches)
        except APIError:
            # Don't mind if some errors happen while downloading the matches
            continue
        except urllib.error.URLError:
            continue

    return runtime_config


from tblib import pickling_support
import pickle

pickling_support.install()

def exc2pickle(exc_info):
    return pickle.dumps(exc_info)

def pickle2exc(pickled_exception):
    type, instance, traceback = pickle.loads(pickled_exception)
    return instance.with_traceback(traceback)

def task_done(queue):
    task_done = getattr(queue, 'task_done', None)
    task_done and task_done()

def close_queue(*queues):
    # Close the queues if they support it
    for queue in queues:
        if hasattr(queue, 'close'):
            queue.close()

    # Join the thread writing to the queue int he background if they have it
    for queue in queues:
        if hasattr(queue, 'join_thread'):
            queue.join_thread()


import os

from cassiopeia.baseriotapi import get_league_entries_by_summoner, get_summoners_by_name, get_versions
from cassiopeia.baseriotapi import get_challenger, get_master, get_match_list
from cassiopeia.baseriotapi import get_match_list, get_match
from cassiopeia.type.api.exception import APIError

if os.environ.get('LOL_SCRAPER_TEST', False):
    from tests.dummy import get_match, get_match_list


import queue
import os

if os.environ.get('LOL_SCRAPER_TEST', False):
    from tests.dummy import get_tier_from_participants
else:
    from lol_scraper.summoners_api import get_tier_from_participants

from lol_scraper.data_types import Maps, Tier, Queue, milliseconds_unix_time
from lol_scraper.summoners_api import check_minimum_patch
from lol_scraper.cassiopeia_proxy import get_match, get_match_list


class FetchingException(Exception):
    def __init__(self, match, msg):
        self.match = match
        self.short_message = msg

    def __repr__(self):
        return "Exception while fetching match {} due to '{}'".format(self.match, self.short_message)

    def __str__(self):
        return self.__repr__()


def fetch(fetch_function, send_result, conf, source, destination, previous, logger, exit_request):
    """

    :param dict                     conf:
    :param queue.Queue              source:
    :param Any                      destination:
    :param ConcurrentSet            previous
    :param DistributedLogger        logger:
    :param multiprocessing.Value    exit_request:
    :return:
    """
    while not exit_request.value:
        try:
            # Poll with a timeout so we can check for the exit condition
            try:
                # Task done is called by the QueueMultiplexer if it's needed
                item_id = source.get(timeout=1)
            except queue.Empty:
                continue

            with previous:
                if item_id in previous:
                    # Item already analyzed
                    continue
                else:
                    # Put it now, so other workers know they shouldn't download it
                    previous.add(item_id)

            try:
                value = fetch_function(item_id, conf)
            except:
                logger.exception()
                continue

            send_result(value, destination)
        except:
            logger.exception()


def _fetch_match(match_id, conf):
    try:
        match = get_match(match_id, conf['include_timeline'])
        if match.mapId == Maps[conf['map_type']].value:
            match_min_tier, participant_tiers = get_tier_from_participants(match.participantIdentities,
                                                                           Tier.parse(conf['minimum_tier']),
                                                                           Queue[conf['queue']])

            valid = (match_min_tier.is_better_or_equal(Tier.parse(conf['minimum_tier'])) and
                     check_minimum_patch(match.matchVersion, conf['minimum_patch']))
            return match, match_min_tier if valid else None, participant_tiers
        else:
            return None, None, {}
    except Exception as e:
        raise FetchingException(match_id, str(e)) from e


def _send_match_results(match_info, destination):
    match, tier, participant_tiers = match_info
    players_to_analyze, downloaded_matches = destination

    for player_id_set in participant_tiers.values():
        for player_id in player_id_set:
            try:
                # We don't care if the queue is full, that means we have enough players to analyze.
                players_to_analyze.put(player_id, block=False)
            except queue.Full:
                # Discard the player and all the subsequent ones
                break

    if tier:
        try:
            # The point of downloading matches is to analyze them. If the queue is full, sleep for a long time (5s) as
            # additional matches are not needed
            downloaded_matches.put((match, tier), timeout=1)
        except queue.Full:
            # Discard the match
            pass


def download_matches(conf, matches_to_download, players_to_analyze, downloaded_matches, past_matches, logger,
                     exit_request):
    """

    :param dict                     conf:
    :param queue.Queue              matches_to_download:
    :param queue.Queue              players_to_analyze:
    :param queue.Queue              downloaded_matches:
    :param ConcurrentSet            past_matches:
    :param DistributedLogger        logger:
    :param multiprocessing.Value    exit_request:
    :return:
    """
    try:
        return fetch(_fetch_match, _send_match_results, conf, matches_to_download,
                     (players_to_analyze, downloaded_matches),
                     past_matches, logger, exit_request)
    finally:
        logger.debug("Match download worker terminating")


def _fetch_player(player_id, configuration):
    match_list = get_match_list(player_id, begin_time=milliseconds_unix_time(configuration['start']),
                                end_time=milliseconds_unix_time(configuration['end']),
                                ranked_queues=configuration['queue'])
    return (match.matchId for match in match_list.matches)


def _send_player_results(match_list, matches_to_download):
    for match_id in match_list:
        try:
            # The only point of downloading players is to provide matches to download. If the queue is already full.
            # sleep for a long time (5s) as the data is not needed, we already have enough to process
            matches_to_download.put(match_id, timeout=1)
        except queue.Full:
            # Discard the processed element and all the subsequent ones
            break


def download_players(conf, players_to_analyze, matches_to_download, past_players, logger, exit_request):
    """

    :param dict                     conf:
    :param queue.Queue              players_to_analyze:
    :param queue.Queue              matches_to_download:
    :param ConcurrentSet            past_players
    :param DistributedLogger        logger:
    :param multiprocessing.Value    exit_request:
    :return:
    """
    try:
        return fetch(_fetch_player, _send_player_results, conf, players_to_analyze, matches_to_download, past_players,
                     logger, exit_request)
    finally:
        logger.debug("Player download worker terminating")


def function_executor(function, match_queue, thread_init_function, thread_index, user_data, logger, exit_request):
    try:
        thread_data = thread_init_function(thread_index, user_data)
    except Exception as e:
        logger.exception()
        raise

    while not exit_request.value:
        try:
            # task_done is called by the FallbackQueue
            match, tier = match_queue.get(timeout=1)
        except queue.Empty:
            continue
        try:
            function(match, tier, thread_data, logger)
        except Exception as e:
            logger.exception()

    logger.debug("User function worker terminating")



import logging
import multiprocessing as mp
import os
import queue
import threading
import time


import cloudpickle
from cassiopeia import baseriotapi

from lol_scraper.concurrency import task, common, process_controller as ctrl
from lol_scraper.data_types import DistributedLogger, Bunch, MessagePipe

max_players_download_threads = int(os.environ.get('MAX_PLAYERS_DOWNLOAD_THREADS', 10))
matches_download_threads = int(os.environ.get('MATCHES_DOWNLOAD_THREADS', 10))
logging_interval = int(os.environ.get('LOGGING_INTERVAL', 60))

MAX_MULTIPROCESS_QUEUE_SIZE = 50

SINGLE_THREAD = 0
MULTI_THREAD = 1
#SINGLE_PROCESS = 2
MULTI_PROCESS = 3


def no_op(*args, **kwargs):
    pass


def setup_riot_api(cassiopeia_conf):
    baseriotapi.set_api_key(cassiopeia_conf['api_key'])
    baseriotapi.set_region(cassiopeia_conf['region'])

    limits = cassiopeia_conf.get('rate_limits', None)
    if limits is not None:
        if isinstance(limits[0], (list, tuple)):
            baseriotapi.set_rate_limits(*limits)
        else:
            baseriotapi.set_rate_limit(limits[0], limits[1])
    baseriotapi.print_calls(cassiopeia_conf.get('print_calls', False))


def _consume_elements(*operation_target_source):
    for operation, target, source in operation_target_source:
        while True:
            try:
                operation(target, source.get(block=False))
                common.task_done(source)
            except queue.Empty:
                break


def _remove_unprocessed_matches(target, item):
    match, tier = item
    target.remove(match.matchId)


def _prepare_initialization_function(conf, number_of_processes, user_initialization_function):
    cassiopeia_conf = {k: v for k, v in conf["cassiopeia"].items()}
    # No point in printing from other processes, as you can not read the stdout
    cassiopeia_conf.pop('print_calls', None)
    # Evenly divide the call limits between processes
    conf_limits = cassiopeia_conf.get('rate_limits', ((10, 10), (500, 600)))
    if isinstance(conf_limits[0], (list, tuple)):
        cassiopeia_conf['rate_limits'] = [(calls // number_of_processes, duration) for calls, duration in conf_limits]
    else:
        cassiopeia_conf['rate_limits'] = [(conf_limits[0] // number_of_processes, conf_limits[1])]

    def initialization_function(bucket_id, configuration):
        setup_riot_api(cassiopeia_conf)
        return user_initialization_function(bucket_id, number_of_processes, configuration)

    return initialization_function


def download_matches(conf, match_downloaded_callback, on_exit_callback=no_op, process_init_function=no_op,
                     thread_init_function=no_op, parallelism=SINGLE_THREAD, threads=None):
    """
    Fetches matches from the Riot API.
    It divides the matches and the players into buckets based on their ID and a partition function (the module operation
    over the ID). Every process handles one bucket, there are multiple threads per process to minimize the impact of
    latency over the download speed.
    A callback is executed every time a match is downlaoded


    :param conf:                        The configuration file specifying the fetching options

    :param match_downloaded_callback:   A function which is executed every time a match is downloaded. Takes as
                                        (the MatchDetail, the match tier, the thread data, the logger) as arguments.
                                        To see what the thread data is see thread_init_function

    :param on_exit_callback:            A function which is executed at the end of the collection process (after an exit
                                        request or in case of error).
                                        Takes (players yet to be analyzed, players already analyzed, matches yet to
                                        download, matches already downloaded). Defaults to no_op

    :param process_init_function:       A function which is executed in every process, before the worker threads are
                                        spawned. Takes (bucket_id, conf) as parameters, returns an object. Defaults to
                                        np_op

    :param thread_init_function:        A function which is executed in every thread, before the worker start it's
                                        operations. Takes (thread_index, user data), where user data is the object
                                        returned by process_init_function. Defaults to no_op

    :param parallelism:                 Specify the parallelism level of the user function. It can be SINGLE_THREAD,
                                        MULTI_THREAD or MULTI_PROCESS.
                                        SINGLE_THREAD:  execute the match_downloaded_callback in a thread in the main
                                                        process.

                                        MULTI_THREAD:   execute the match_download_callback in multiple threads in the
                                                        main process.

                                        MULTI_PROCESS:  execute the match_download_callback in multiple threads in all
                                                        the worker processes

    :param threads:                     A tuple (number of player download workers, number of match download workers,
                                        number of callback execution workers) specifying for each process how many
                                        threads should be spawned.
                                        NOTE: it's usually best to keep the match workers double of the player workers,
                                        as the match worker calls the Riot API 2 times for every match, while the
                                        player worker calls it once for every player.
    """

    NUMBER_OF_PROCESSES = int(os.environ.get('PROCESSES', 1))
    _underlying_logger = logging.getLogger("distributed_logger")
    # Defaults to SINGLE_THREAD, eg, 0 user function workers in the other processes
    if threads is None:
        threads = [1,2,0]
    else:
        threads = list(threads)

    if conf['logging_level'] != logging.NOTSET:
        logging_level = conf['logging_level']
        _underlying_logger.setLevel(logging_level)
    else:
        logging_level = _underlying_logger.getEffectiveLevel()

    def checkpoint(players_to_analyze, analyzed_players, matches_to_download, downloaded_matches):
        if on_exit_callback and on_exit_callback != no_op:
            try:
                on_exit_callback(players_to_analyze, analyzed_players, matches_to_download, downloaded_matches)
            except:
                _underlying_logger.exception("Exception while executing the checkpoint callback")
                raise
        _underlying_logger.info("Checkpoint callback executed successfully")

    next_players = set(conf['seed_players_id'])
    next_matches = set(conf['matches_to_download'])
    past_matches = set(conf['downloaded_matches'])
    past_players = set()

    if len(next_matches) == 0 and len(next_players) == 0:
        raise ValueError("You need to provide at least a match or a player to start fetching from")

    _underlying_logger.info("{} previously downloaded matches".format(len(past_matches)))
    _underlying_logger.info("{} matches to download".format(len(next_matches)))

    assert NUMBER_OF_PROCESSES > 0
    logger_thread = None
    logger_exit_request = Bunch(value=False)
    queue_type = mp.Queue if NUMBER_OF_PROCESSES > 1 else queue.Queue

    # The logger should never discard messages, so the queue is unbounded
    logging_queue = queue_type()

    # For the local worker use a thread queue, not a multiprocess one
    worker_connections, remote_connections = [], []
    for _ in range(NUMBER_OF_PROCESSES):
        conn1, conn2 = mp.Pipe()
        worker_connections.append(MessagePipe(conn1))
        remote_connections.append(MessagePipe(conn2))

    player_queues = [queue_type(maxsize=MAX_MULTIPROCESS_QUEUE_SIZE) for _ in range(NUMBER_OF_PROCESSES)]
    match_queues = [queue_type(maxsize=MAX_MULTIPROCESS_QUEUE_SIZE) for _ in range(NUMBER_OF_PROCESSES)]

    downloaded_matches = queue_type(maxsize=MAX_MULTIPROCESS_QUEUE_SIZE)
    local_bucket_id = 0

    processes = []
    logger = DistributedLogger(logging_level, logging_queue)
    try:
        logger_thread = threading.Thread(name="Logger",
                                         target=task.log_messages, args=(_underlying_logger, logging_queue, logger_exit_request))
        logger_thread.start()

        # Initialize the queues with the players and matches to download
        for player_id in next_players:
            _bucket_id = player_id % NUMBER_OF_PROCESSES
            try:
                player_queues[_bucket_id].put(player_id, block=False)
            except queue.Full:
                logger.warn("Ignoring some initial players (queue full)")
                break

        for match_id in next_matches:
            _bucket_id = match_id % NUMBER_OF_PROCESSES
            try:
                match_queues[_bucket_id].put(match_id, block=False)
            except queue.Full:
                logger.warn("Ignoring some initial matches (queue full)")
                break

        for _bucket_id, connection in enumerate(worker_connections):
            connection.put(ctrl.make_message(ctrl.PAST_PLAYERS_AND_MATCHES, _bucket_id,
                                             (past_players, past_matches)))

        # Since the execution must be single threaded, in the main process, the other processes should have0 0 user
        # function threads
        if parallelism in (SINGLE_THREAD, MULTI_THREAD):
            threads[2] = 0
        pickled_process_init = cloudpickle.dumps(_prepare_initialization_function(conf, NUMBER_OF_PROCESSES,
                                                                                  process_init_function))
        pickled_thread_init = cloudpickle.dumps(thread_init_function)
        pickled_user_function = cloudpickle.dumps(match_downloaded_callback)
        _bucket_id = 1
        for connection in remote_connections[1:]:
            p = mp.Process(name="ProcessBucket%s" % _bucket_id, target=ctrl.download_and_execute,
                           args=(conf, _bucket_id, player_queues, match_queues, downloaded_matches,
                                 connection, pickled_user_function, pickled_thread_init, logging_level,
                                 logging_queue, pickled_process_init, threads))
            processes.append(p)
            p.start()
            _bucket_id += 1

        # If the execution should be single threaded, we want 1 thread in the main process. If it's multi-threaded
        # we want the user to pick the number of user processes, but there should always be at least 1
        if threads[2] == 0  and parallelism in (MULTI_THREAD, MULTI_PROCESS):
            logger.warn("Execution set as MULTI_PROCESS or MULTI_THREAD, but 0 user threads per process were set. "
                        "Falling back to SINGLE_THREAD")
        threads[2] = max(1, threads[2])
        # Start a worker in this process as well
        local_worker = threading.Thread(name="ThreadBucket0", target=ctrl.download_and_execute,
                                        args=(conf, local_bucket_id, player_queues, match_queues, downloaded_matches,
                                              remote_connections[local_bucket_id], pickled_user_function,
                                              pickled_thread_init, logging_level, logging_queue,
                                              pickled_process_init, threads))
        processes.append(local_worker)
        local_worker.start()

        logger.info("Starting fetching..")

        # Send an ID when you make requests, so that you know to which request the response is
        monitoring_request_id = 0
        i = 0
        while not conf.get('exit', False):
            # Execute every LOGGING_INTERVAL seconds
            if i % logging_interval == 0:
                for _bucket_id, connection in enumerate(worker_connections):
                    connection.put(ctrl.make_message(ctrl.MONITORING_REQUEST, _bucket_id, monitoring_request_id))

                waiting_players = 0
                completed_players = 0
                waiting_matches = 0
                completed_matches = 0
                matches_to_process = 0

                # Here we send to the bucket a request to communicate us several statistics over their performances
                # We try 3 times to get the data from them, then we write the message ignoring the missing buckets
                not_replying_buckets = {x for x in range(len(worker_connections))}
                _retries = 3
                while not_replying_buckets and _retries:
                    _retries -= 1
                    for index, connection in enumerate(worker_connections):
                        if index in not_replying_buckets:
                            # We are still waiting for a response
                            try:
                                message = connection.get(timeout=0.5)
                            except queue.Empty:
                                pass
                            else:
                                type = ctrl.get_type(message)
                                if type == ctrl.MONITORING_RESPONSE:
                                    monitoring_response_id, payload = ctrl.get_payload(message)

                                    if monitoring_response_id != monitoring_request_id:
                                        logger.debug("Received a response id:%i instead of %i from bucket %i",
                                                     monitoring_response_id, monitoring_request_id, index)
                                        # Skip the old message
                                        continue

                                    w_players, c_players, w_matches, c_matches, l_process = payload
                                    waiting_players += w_players
                                    completed_players += c_players
                                    waiting_matches += w_matches
                                    completed_matches += c_matches
                                    matches_to_process += l_process
                                    not_replying_buckets.remove(index)
                                elif type == ctrl.INITIATING_SHUTDOWN:
                                    logger.info("Bucket %s terminated unexpectedly", index)
                                    return
                                else:
                                    raise ValueError("Unknown message: {}".format(message))

                _msg = ("Players in queue: %s. Players downloaded: %s. Matches in queue: %s. Matches downloaded: %s. "
                        "Matches waiting to be processed: %s")
                _log_args = [waiting_players, completed_players, waiting_matches, completed_matches,
                             matches_to_process + downloaded_matches.qsize()]

                # If not all buckets reported, add it to the message
                if not_replying_buckets:
                    _msg += ". Bucked which did not report: %s"
                    _log_args.append(", ".join(map(str, not_replying_buckets)))

                logger.info(_msg, *_log_args)

            i += 1
            time.sleep(1)

        logger.info("Terminating fetching")

    except KeyboardInterrupt:
        logger.info("user requested for exit")
        return
    except:
        _underlying_logger.exception("Exception in the main")
        raise
    finally:
        conf['exit'] = True
        logger.info("Initiating termination")
        logger.info("Sending the exit signal to the workers")
        for _bucket_id, connection in enumerate(worker_connections):
            try:
                connection.put(ctrl.make_message(ctrl.EXIT_REQUEST, _bucket_id))
            except queue.Full:
                # In the current implementation, connection is actually a Pipe connection. When the other process
                # terminates it closes the pipe. Sending an EXIT_REQUEST to a terminated process will thus result in a
                # BrokenPipeError, which gets wrapped as a queue.Full by the adapter (yeah, maybe not the best choice).
                # The process, if it terminated cleanly (it was not killed) will send all the information to the pipe
                # before exiting, so we already have the information we need. Just skip this error
                logger.debug("Bucket %s already terminated. Failed to send EXIT_REQUEST", _bucket_id)

        logger.info("Retrieving the data from the workers")
        past_players = set()
        past_matches = set()
        # Retrieve all the players who have been downloaded by the workers
        for connection in worker_connections:
            while True:
                # TODO: what if one process failed (and will never send a message)? Implement a retry mechanic
                message = connection.get_ignore(ctrl.msg_ignore(ctrl.INITIATING_SHUTDOWN, ctrl.MONITORING_RESPONSE))
                type = ctrl.get_type(message)
                if type == ctrl.PAST_PLAYERS_AND_MATCHES:
                    remote_past_players, remote_past_matches = ctrl.get_payload(message)
                    past_players.update(remote_past_players)
                    past_matches.update(remote_past_matches)
                    break
                else:
                    raise ValueError("Unexpected message: {}".format(message))

        logger.info("Consuming the queues and joining the workers")
        # We need to clear the queue where the process is writing in order to allow it to terminate
        next_players = set()
        next_matches = set()
        # Wait for all the processes to exit, consuming elements from the queue where they write to
        reported = []
        while len(reported) != len(worker_connections):
            for ind, (connection, player_q, match_q) in enumerate(zip(worker_connections, player_queues, match_queues)):
                # Skip the worker which already replied
                if ind in reported:
                    continue
                # Consume all the available items in the queue, so that the process has space to write them
                _consume_elements((set.add, next_players, player_q),
                                  (set.add, next_matches, match_q),
                                  (_remove_unprocessed_matches, past_matches, downloaded_matches))
                logger.debug("Consumed the queues for bucket %s", ind)
                # Check if the process has terminated
                try:
                    message = connection.get_ignore(ctrl.msg_ignore(ctrl.INITIATING_SHUTDOWN, ctrl.MONITORING_RESPONSE),
                                                    timeout=0.5)
                except queue.Empty:
                    pass
                else:
                    type = ctrl.get_type(message)
                    if type == ctrl.SHUTDOWN_COMPLETED:
                        reported.append(ind)
                        logger.debug("Bucket %s joined", ind)
                    else:
                        raise ValueError("Unexpected message: {}".format(message))

        # After all the processes terminated, clean a last time the queues
        for player_q, match_q in zip(player_queues, match_queues):
            _consume_elements((set.add, next_players, player_q),
                              (set.add, next_matches, match_q),
                              (_remove_unprocessed_matches, past_matches, downloaded_matches))

        # We don't need the queues anymore. Only leave the logger queue open
        # The queues must be empty
        queues_to_close = match_queues + player_queues + [downloaded_matches]
        assert all(not q.qsize() for q in queues_to_close), str([q.qsize() for q in queues_to_close])

        logger.info("Closing the queues")
        common.close_queue(*queues_to_close)

        logger.info("Joining the workers")

        # Joining workers once the queues are empty
        for worker in processes:
            worker.join()

        # At this point all workers terminate and it's not possible for them to generate new messages.
        # Terminate the logger
        logger.info("Terminating the distributed log writer")
        logger_exit_request.value = True
        logger_thread.join()

        # Always call the checkpoint, so that we can resume the download in case of exceptions.
        _underlying_logger.info("Calling checkpoint callback")
        checkpoint(next_players, past_players, next_matches, past_matches)
        _underlying_logger.info("Bye bye")



import os
from lol_scraper.data_types import Tier
import random

class Bunch:
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


match_file = os.path.join(os.path.dirname(__file__), 'match_string.json')
with open(match_file, 'rt') as f:
        match_string = f.read()
        from cassiopeia.type.dto.match import MatchDetail
        import json

import itertools, random
count = itertools.count(random.randint(0, 1000000))
sleep_time = 0.10
from collections import defaultdict
thread_id = itertools.count()
thread_names = defaultdict(lambda : next(thread_id))
exception_rate = 0#0.01

def get_match_list(*args, **kwargs):
    from cassiopeia.dto import requests
    import time
    time.sleep(sleep_time)
    requests.rate_limiter.call()
    obj = Bunch(matches=[Bunch(matchId=next(count)) for _ in range(random.choice([0,1,2,3,4]))])
    if random.random() < exception_rate:
        raise ValueError()
    return obj

def get_match(matchId, include_timeline=True):
    from cassiopeia.dto import requests
    import time
    time.sleep(sleep_time)
    requests.rate_limiter.call()

    match = MatchDetail(json.loads(match_string))
    match.matchId = matchId
    if random.random() < exception_rate:
        raise ValueError()
    return match

def get_tier_from_participants(*args, **kwargs):
    from cassiopeia.dto import requests
    import time
    requests.rate_limiter.call()
    time.sleep(sleep_time)
    return Tier.challenger, {Tier.challenger: {next(count) for _ in range(10)}}


import logging
import os

# Load the dummy implementation of the network calls
os.environ['LOL_SCRAPER_TEST'] = "1"

from datetime import datetime as dt

from lol_scraper.match_downloader import download_matches, SINGLE_THREAD
from lol_scraper.initialize_configuration import prepare_config


conf = {
     "cassiopeia": {
        "api_key": "",
         "rate_limits": (3000, 10),
        "region": "KR"
    },
    "include_timeline": False,
    "seed_players_id" : [-10,-9,-8,-7,-6,-5,-4,-3,-2,-1]
}

class Container(): pass

def checkpoint(next_players, past_players, next_matches, past_matches):
    global ret
    ret = (next_players, past_players, next_matches, past_matches)

def check_perf(sleep_seconds=0, limits=((10,10),(500, 600))):
    def initialize(bucket_id, number_of_processes, conf):
        import tests.dummy
        tests.dummy.sleep_time = sleep_seconds

    store = Container()
    store.total_calls = 0
    store.start = None
    store.last = None
    store.partial_calls = 0
    store.avg = 0

    runtime_conf = prepare_config(conf)
    runtime_conf['cassiopeia']['rate_limits'] = limits

    def callback(match, tier, data, logger):
        now = dt.now()
        if store.start is None:
            store.start = now
        if store.last is None:
            store.last = now

        if store.total_calls % 10 == 0:
            elapsed_time = max(0.01, (now - store.last).total_seconds())
            total_time = max(0.01, (now - store.start).total_seconds())
            immediate = store.partial_calls / elapsed_time
            store.avg = (store.avg + immediate)/2
            logger.info("calls:%i  tot time:%i spd:%.2f avg spd:%.2f",
                        store.total_calls, total_time, store.total_calls/total_time, store.avg)
            store.last = now
            store.partial_calls = 0

        store.total_calls += 1
        store.partial_calls += 1

    logging.basicConfig(format='%(asctime)s, %(levelname)s, %(message)s',
                        datefmt="%m-%d %H:%M:%S",
                        level=logging.INFO)

    download_matches(callback, checkpoint, runtime_conf, parallelism=SINGLE_THREAD,
                     process_init_function=initialize)

if __name__ == '__main__':
    check_perf()


import operator
import threading
from collections import defaultdict

from lol_scraper.data_types import Tier, Queue, SimpleCache, cache_autostore
from lol_scraper.cassiopeia_proxy import get_league_entries_by_summoner, get_summoners_by_name, get_versions

LATEST = "latest"
version_key = 'current_version'
_cache = SimpleCache()

_patch_changed_lock = threading.Lock()
_patch_changed = False


def _slice(start, stop, step):
    """
    Generate pairs so that you can slice from start to stop, step elements at a time
    :param start: The start of the generated series
    :param stop: The last of the generated series
    :param step: The difference between the first element of the returned pair and the second
    :return: A pair that you can use to slice
    """
    if step == 0:
        raise ValueError("slice() arg 3 must not be zero")
    if start == stop:
        raise StopIteration

    _previous = start
    _next = start + step
    while _next < stop:
        yield _previous, _next
        _previous += step
        _next += step
    yield _previous, stop


def leagues_by_summoner_ids(summoner_ids, queue=Queue.RANKED_SOLO_5x5):
    """
    Takes in a list of players ids and divide them by league tiers.
    :param summoner_ids: a list containing the ids of players
    :param queue: the queue to consider
    :return: a dictionary tier -> set of ids
    """
    summoners_league = defaultdict(set)
    for start, end in _slice(0, len(summoner_ids), 10):
        for player_id, leagues in get_league_entries_by_summoner(summoner_ids[start:end]).items():
            for league in leagues:
                if Queue[league.queue] == queue:
                    summoners_league[Tier.parse(league.tier)].add(int(player_id))
    return summoners_league


def get_tier_from_participants(participants_identities, minimum_tier=Tier.bronze, queue=Queue.RANKED_SOLO_5x5):
    """
    Returns the tier of the lowest tier and the participantsIDs divided by tier
    player in the match
    :param participants_identities: the match participants
    :param minimum_tier: the minimum tier that a participant must be in order to be added
    :param queue: the queue over which the tier of the player is considered
    :return: the tier of the lowest tier player in the match
    """
    leagues = leagues_by_summoner_ids([p.player.summonerId for p in participants_identities], queue)
    match_tier = max(leagues.keys(), key=operator.attrgetter('value'))
    return match_tier, {league: ids for league, ids in leagues.items() if league.is_better_or_equal(minimum_tier)}


def summoner_names_to_id(summoners):
    """
    Gets a list of summoners names and return a dictionary mapping the player name to his/her summoner id
    :param summoners: a list of player names
    :return: a dictionary name -> id
    """
    ids = {}
    for start, end in _slice(0, len(summoners), 40):
        result = get_summoners_by_name(summoners[start:end])
        for name, summoner in result.items():
            ids[name] = summoner.id
    return ids


def set_patch_changed(*_, **__):
    with _patch_changed_lock:
        global _patch_changed
        _patch_changed = True


def consume_path_changed():
    with _patch_changed_lock:
        global _patch_changed
        _patch_changed = False


def get_patch_changed():
    with _patch_changed_lock:
        global _patch_changed
        return _patch_changed


@cache_autostore(version_key, 60 * 60, _cache, on_change=set_patch_changed)
def get_last_patch_version():
    version_extended = get_versions()[0]
    version = ".".join(version_extended.split(".")[:2])
    return version


def check_minimum_patch(patch, minimum):
    if not minimum:
        return True
    if minimum.lower() != LATEST:
        return patch >= minimum
    else:
        try:
            version = get_last_patch_version()
            return patch >= version
        except Exception:
            # in case the connection failed, do not store it, and try next round
            # Reject every version as we are not sure which is the latest version
            # and we don't want to pollute the data with patches with the wrong version
            return False



#!/usr/bin/env python

from setuptools import setup, find_packages


install_requires = [
    "cassiopeia",
    "cloudpickle",
    "tblib"
]

project_url = "https://github.com/MakersF/LoLScraper"
setup(
    name="lol_scraper",
    version="0.2.0",
    author="Francesco Zoffoli",
    author_email="makers.f.dev@gmail.com",
    url=project_url,
    description="A python script and library to download and store League of Legends matches with Riot API",
    long_description="Head to the github project page ({}) for a formatted description".format(project_url),
    keywords=["LoL", "League of Legends", "Riot Games", "API", "REST"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Games/Entertainment",
        "Topic :: Games/Entertainment :: Real Time Strategy",
        "Topic :: Games/Entertainment :: Role-Playing",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    license="MIT",
    packages=find_packages(),
    zip_safe=True,
    install_requires=install_requires
)